import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintStream;
import java.time.Duration;
import java.util.Scanner;

import static org.junit.Assert.assertEquals;
import static org.junit.jupiter.api.Assertions.assertTimeoutPreemptively;

import org.junit.jupiter.api.Test;

public class InterpreterTests {

    String getExpectedOutput(File file) {
        String expected = "";
        try {
            Scanner scanner = new Scanner(file);
            String line = scanner.nextLine();
            while (!line.equals("% expected output:")) {
                line = scanner.nextLine();
            }
            while (scanner.hasNextLine()) {
                expected += scanner.nextLine().substring(2) + "\n"; // trim off "% "
            }
            scanner.close();
        } catch (FileNotFoundException ex) {
            System.out.println(file.getName() + " doesn't exist");
        }
        return expected;
    }

    String runTest(String filename) {
        // redirect stdout to
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        PrintStream ps = new PrintStream(baos);
        PrintStream oldOut = System.out;
        System.setOut(ps);

        Reader r = new Reader("tests/" + filename);
        try {
            Interpreter interpreter = new Interpreter();
            Token t;
            while (r.hasNext()) {
                t = (Token) r.next();
                if (t.isSymbol() && // only if symbol:
                        t.getSymbol().equals("quit"))
                    break;
                // process token
                interpreter.processToken(t);
            }
        } catch (Exception ex) {
            System.out.println(filename + " FAILED due to exception:");
            System.out.println(ex);
        }

        // reinstate stdout, print test results
        System.out.flush();
        System.setOut(oldOut);
        String actual = baos.toString().replaceAll("\r", "");
        return actual;
    }

    void checkOutput(String expected, String actual, String testName) {
        if (expected.equals(actual)) {
            System.out.println(testName + " PASSED");
        } else {
            System.out.println(testName + " FAILED");
            System.out.println("Expected");
            System.out.println(expected);
            System.out.println("but got");
            System.out.println(actual);
        }
        assertEquals(expected, actual);
    }

    @Test
    void testArithmetic() {
        assertTimeoutPreemptively(Duration.ofSeconds(1), () -> {
            File file = new File("tests/arithmetic-test.ps");
            checkOutput(getExpectedOutput(file), runTest(file.getName()), file.getName());
        });
    }

    @Test
    void testBoolean() {
        assertTimeoutPreemptively(Duration.ofSeconds(1), () -> {
            File file = new File("tests/boolean-test.ps");
            checkOutput(getExpectedOutput(file), runTest(file.getName()), file.getName());
        });
    }

    @Test
    void testCountTo10() {
        assertTimeoutPreemptively(Duration.ofSeconds(1), () -> {
            File file = new File("tests/count-to-10.ps");
            checkOutput(getExpectedOutput(file), runTest(file.getName()), file.getName());
        });
    }

    @Test
    void testDef() {
        assertTimeoutPreemptively(Duration.ofSeconds(1), () -> {
            File file = new File("tests/def-test.ps");
            checkOutput(getExpectedOutput(file), runTest(file.getName()), file.getName());
        });
    }

    @Test
    void testDup() {
        assertTimeoutPreemptively(Duration.ofSeconds(1), () -> {
            File file = new File("tests/dup-test.ps");
            checkOutput(getExpectedOutput(file), runTest(file.getName()), file.getName());
        });
    }

    @Test
    void testExch() {
        assertTimeoutPreemptively(Duration.ofSeconds(1), () -> {
            File file = new File("tests/exch-test.ps");
            checkOutput(getExpectedOutput(file), runTest(file.getName()), file.getName());
        });
    }

    @Test
    void testFib() {
        assertTimeoutPreemptively(Duration.ofSeconds(1), () -> {
            File file = new File("tests/fib.ps");
            checkOutput(getExpectedOutput(file), runTest(file.getName()), file.getName());
        });
    }

    @Test
    void testPop() {
        assertTimeoutPreemptively(Duration.ofSeconds(1), () -> {
            File file = new File("tests/pop-test.ps");
            checkOutput(getExpectedOutput(file), runTest(file.getName()), file.getName());
        });
    }

    @Test
    void testProc() {
        assertTimeoutPreemptively(Duration.ofSeconds(1), () -> {
            File file = new File("tests/proc-test.ps");
            checkOutput(getExpectedOutput(file), runTest(file.getName()), file.getName());
        });
    }

    @Test
    void testPstack() {
        assertTimeoutPreemptively(Duration.ofSeconds(1), () -> {
            File file = new File("tests/pstack-test.ps");
            checkOutput(getExpectedOutput(file), runTest(file.getName()), file.getName());
        });
    }
}
